class BaseDispatcher:

  "Base Abstract Class for Dispatchers"


  def __init__(self,):
    """
      Initializes Dispatchers and starts tasks queues
    """
    pass
  
  def shutdown(self,):
    """
      shutdowns subprocesses and releases memory being used by dispatcher.
    """
    raise NotImplementedError

  def run(self, func, *args, **kwargs):
    """
      Places a task into the servers queued tasks for client completion.

      # Returns 
      The associated task id to recover the results.
    """
    raise NotImplementedError
  
  def join(self,):
    """
      Gets results from all queued tasks and subsequently shuts down
      the dispatcher.
    """
    results = self.get_results(task_ids=None, values_only=True)
    self.shutdown()
    return results
  
  def get_results(self, task_ids=None, values_only=True):
    """
      Gets the results to tasks associated with the passed in task_ids. 
      Hangs current thread until all the tasks are complete.

      task_ids: task_ids as generated by self.run(). These are used by the 
        server to identify which task to return the results for.
      values_only: if False returns dictionary that includes the task ids with its 
        results. Otherwise, returns the values computed in order of the 
        requested task ids.
    """
    raise NotImplementedError



import datetime
from threading import Thread
from multiprocessing import Pipe, Process
from .utils import Packet
from .base import BaseDispatcher

class MulticoreDispatcher(BaseDispatcher):

  def __init__(self, workers=1, timeout=None):
    """
      Initializes a MulticoreDispatcher with a configurable number of cores
      
      #Arguments
      workers: Int. Max number of worker subprocesses to spawn.
      timeout: Max time in seconds to permit a Process to execute any given task.
        If None, the processes will continue to run until self.shutdown() is called.
    """
    

    self._close_thread   = False       #If True, monitor thread will close
    self.tasks           = {}          #task_id: Task
    self.active          = set()       #task_ids for active tasks
    self.queued          = set()       #task_ids that are waiting to be run
    self.timeout         = timeout     #max time for process
    self.workers         = workers     #max number of worker processes to spawn at one time
    self.current_task_id = 1           #Manages task ids, tasks start from 1
    

    self.processes = {} #process_id: {'connections', 'process', 'running'}
    for p_id in range(self.workers):
      parent_conn, sub_conn = Pipe(duplex=True)
      p = Process(
        target=MulticoreDispatcher._spawn_subprocess, 
        args=(sub_conn,),
      )
      p.start()
      self.processes[p_id] = {
        'connections': (parent_conn, sub_conn),
        'process': p,
        'running': False
      }
    
    self.monitor_thread = Thread(
      target=MulticoreDispatcher._monitor,
      args=(self,)
    )
    self.monitor_thread.daemon = True
    self.monitor_thread.start()

  @staticmethod
  def _monitor(self):
    """
      Asynchronously monitors the dispatcher and performs the following:
      1. Assigns idled processes any available queued tasks
      2. Check each active tasks to ensure timeouts have not been reached
      3. Replaces crashed subprocesses with new subprocesses
      4. Receives task results from subprocesses
    """
    while not self._close_thread:
      
      ####### Check each active task for timeout or subprocess crash #######
      for task_id in list(self.active):
        #timeout
        task = self.tasks[task_id]
        timeout = task['timeout']
        dt = (datetime.datetime.now() - task['start_time']).total_seconds()
        timeout_reached = dt > timeout if timeout is not None else False
        
        #subprocess crashed
        p_id = task['p_id']
        pfield = self.processes[p_id]
        process = pfield['process']
        process_crashed = process.exitcode is not None

        if timeout_reached or process_crashed:
          #Kill attached process
          process.terminate()

          #Create new subprocess
          parent_conn, sub_conn = pfield['connections']
          new_process = Process(
            target=MulticoreDispatcher._spawn_subprocess, 
            args=(sub_conn,),
          )
          new_process.start()

          #Attach connections to new subprocess
          pfield.update({
            'connections': (parent_conn, sub_conn),
            'process': new_process,
            'running': False
          })

          #Mark task as completed
          task.update({
            'p_id': None,
            'running': False,
            'completed': True,
            'terminated_early': True
          })
          self.active.remove(task_id)


      ####### Check for completed tasks #######
      for p_id in self.processes.keys():
        pfield = self.processes[p_id]
        parent_conn = pfield['connections'][0]
        has_data = parent_conn.poll()
        if has_data:
          #Get response to task from subprocess
          packet = parent_conn.recv()
          response = packet.unpack()

          task_id = response['task_id']
          result  = response['result']

          task = self.tasks[task_id]
          task.update({
            'p_id': None,
            'result': result,
            'running': False,
            'completed': True,
          })
          self.active.remove(task_id)
          pfield['running'] = False


      ######## Get idle processes #######
      idle_processes = []
      for i in self.processes.keys():
        if not self.processes[i]['running']:
          idle_processes.append(i)

      ####### Assign task to any idle process #######
      p_index = len(idle_processes)-1
      while p_index >= 0:
        if len(self.queued):
          task_id = self.queued.pop()
          self.active.add(task_id)
          task = self.tasks[task_id]
          task['p_id'] = idle_processes[p_index]
          task['running'] = True
          task['start_time'] = datetime.datetime.now()

          packet = Packet({
            'task_id': task_id,
            'func': task['func'],
            'args': task['args'],
            'kwargs': task['kwargs'],
          })
          packet.compress(iterations=0)

          pfield = self.processes[idle_processes[p_index]]
          parent_conn = pfield['connections'][0]
          parent_conn.send(packet)

        idle_processes.pop()
        p_index -= 1
  
  @staticmethod
  def _spawn_subprocess(conn):
    """
      Spawns a subprocess that listens to a connection and performs the 
      tasks that are passed through the connection.
    """
    while True:
      #Get task
      packet = conn.recv()
      task = packet.unpack()
      func = task['func']
      args = task['args']
      kwargs = task['kwargs']

      #Perform task
      result = func(*args, **kwargs) #Create a perform task function for abstration???

      #Send reponse
      packet = Packet({
        'task_id': task['task_id'],
        'result': result
      })
      conn.send(packet)

  def _new_task(self,):
    #Task Schema
    task_id = self.current_task_id
    self.current_task_id += 1
    task = {
      'p_id':             None, 
      'start_time':       None, 
      'running':          False, 
      'result:':          None,
      'timeout':          self.timeout,
      'completed':        False,
      'terminated_early': False,
    }
    return task_id, task
  
  def shutdown(self,):
    """
      Shutdown monitor thread, terminates spawned subprocesses, and releases 
      memory allocation resources utilized by the dispatcher. 
    """
    self._close_thread = True
    self.monitor_thread.join()
    ######## Close processes, Free memory #########

    for pfield in self.processes.values():
      process = pfield['process']
      process.terminate()

    del self.tasks

  def run(self, func, *args, **kwargs):
    """
      Places a task into the dispatchers queued tasks for subprocess completion.

       # Arguements
      func: a function
      args: value based arguements for function `func`
      kwargs: keyword based arguements for `func`

      # Returns 
      The associated task id needed to recover the results.
    """
    task_id, task = self._new_task()
    task.update({
      'func': func,
      'args': args,
      'kwargs': kwargs,
    })
    self.queued.add(task_id)
    self.tasks[task_id] = task

    return task_id
  
  def clear_task(self, task_id):
    """
      Removes all traces of a task being present on the server.
      Removes the task_id from all task queues and opens memory for additional 
      tasks.
    """

    for collection in [self.queued, self.active]:
      if task_id in collection:
        collection.remove(task_id)
    
    self.tasks.pop(task_id, None)
  
  def get_results(self, task_ids=None, values_only=True, clear=True):
    """
      Gets the results to tasks associated with the passed in task_ids. 
      Hangs current thread until all the tasks are complete.

      task_ids: list of task_ids as generated by self.run(). These are used by the 
        server to identify which task to return the results for. 
        If `None`, waits for and returns results for all tasks.
      values_only: if False returns dictionary that includes the task ids with its 
        results. Otherwise, returns the values computed in order of the 
        requested task ids.
      clear: If True, removes tasks from dispatcher memory after returning results.
    """
    if task_ids is None:
      task_ids = list(self.tasks.keys())
    
    waiting = True #Continue waiting for tasks to finish
    while waiting:
      
      waiting = False
      #Check that all tasks are completed
      for task_id in task_ids:
        task = self.tasks[task_id]

        if not task['completed']:
          waiting = True

    results = []
    for task_id in task_ids:
      task = self.tasks[task_id]
      
      result = task.get('result')
      results.append(result)
      if clear:
        self.clear_task(task_id)
    
    if values_only:
      retval = results
    else:
      retval = dict(zip(task_ids, results))

    return retval



import datetime
from threading import Thread
from multiprocessing import Lock
from multiprocessing.connection import Listener
from multiprocessing.managers import BaseManager
from .logging import get_logger
from .utils import Packet, Promise, get_free_port

class ParallelManager(BaseManager):  
  def __init__(self, *args, timeout=None, task_limit=None, revive_limit=0, **kwargs):
    """
      Initlizes a parallel manager for distributed task management.

      # Arguments
      address: (string server_ip, int port). 
        Port to open on the server and its IP address for remote connections.
      authkey: authorization key to connect to remote manager.
      timeout: Default time in seconds to permit for a task. 
        If a task takes longer, the server ceases to await a response.
      task_limit: Int. Max number of tasks for server to remember. 
        This monitors the total number of active, completed, and queued tasks.
      revive_limit: Int. Max number of times a client can reach a timeout before 
        it is disconnected from the server.
    """
    super().__init__(*args, **kwargs)

    self.timeout    = timeout
    self.task_limit = task_limit

    self.current_task_id   = 1
    self.current_client_id = 1 #Client id as used by the server to assign tasks
    self.fetch_results_thread_id  = 1
    self.revive_limit = revive_limit #Limit to number of times a timeout can occur on client
    self.clients = {
      #client_id: {
      #  'listener':   mp.Listener,
      #  'connection': mp.connection.Connection,
      #  'busy':   False,
      #  'alive':  True,
      #  'times_dead': 0 #Used to gauge continued revival of this client
      #}
    }
    self.fetch_results = {
      # 'id': {
      #   'thread': Thread,
      #   'listener': mp.Listener
      # }
    }
    self.queued_tasks    = set() #task_ids
    self.active_tasks    = set() #task_ids
    self.completed_tasks = set() #task_ids
    self.tasks = {} #{task_id: Task}... 
    # Task SCHEMA:
    # task = {
    #   'task_id':          task_id,
    #   'func':             func,
    #   'args':             args,
    #   'kwargs':           kwargs,
    #   'start_time':       None, 
    #   'running_on':       None, #client_id 
    #   'completed':        False,
    #   'result':           None,
    #   'timeout':          timeout or self.timeout,
    #   'terminated_early': False
    # }
    
    #Exposed methods to remote clients and drivers (via authkey)
    self.register('submit',      callable=self.submit)
    self.register('monitor',     callable=self.monitor)
    self.register('get_results', callable=self.get_results)
  
  @staticmethod
  def _monitor_clients(self,):
    """
      Runs asynchonously to monitor server. 
      Performs the following functions:
        1. Assigns idled clients any available queued tasks
        2. Check each active tasks to ensure timeouts have not been reached
        3. Marks clients with timeouts as 'dead'; listens for a revival packet
        4. Receives task results from clients
        5. Close results-fetching threads.
    """
    while not self._close_thread:
      ######## Critical Code (Async access to variables prohibited) ########
      self.critical_lock.acquire()

      ######## Check for queued tasks and idle clients ######## 
      if len(self.queued_tasks):
        idle_clients = self._get_idle_clients()

        if len(idle_clients):
          num_tasks_to_assign = min(len(idle_clients), len(self.queued_tasks))
          for i in range(num_tasks_to_assign):
            client_id = idle_clients[i]
            self._send_queued_task(client_id)

      ######## Check for timeouts and mark clients as 'dead' ########
      if len(self.active_tasks):
        tasks_to_kill = set()

        for task_id in self.active_tasks:
          task = self.tasks[task_id]
          start_time   = task.get('start_time')
          max_duration = task.get('timeout')
          duration = (datetime.datetime.now() - start_time).total_seconds()

          if max_duration and duration > max_duration:
            tasks_to_kill.add(task_id)
        
        self._kill_tasks(tasks_to_kill)
      
      ######## Receive task results and check for revival packet ########
      #Use list because popping from self.clients in event of connection error
      client_ids = list(self.clients.keys())
      for client_id in client_ids:
        client = self.clients[client_id]
        server_conn = client['connection']
        
        if server_conn:
          task_done = server_conn.poll()
          if task_done:
            try:
              packet = server_conn.recv()
            except Exception as e:
              msg = 'Exception occurred while reading from client. ' + \
                'Terminating connection.'
              self._kill_client(client_id)
              get_logger().warning(msg)
              continue

            data = packet.unpack()

            result  = data['result']
            task_id = data['task_id']

            if task_id in self.active_tasks:
              self._complete_task(task_id, result)
            elif task_id in self.completed_tasks: #Task was terminated early
              #Do not update results on an already marked `completed` task
              client['alive'] = True
      
      ######## Close results-fetching threads ########
      thread_id_keys = list(self.fetch_results.keys()) #Because popping from self.fetch_results
      for thread_id in thread_id_keys:
        tfield = self.fetch_results[thread_id]
        thread = tfield['thread']
        #   thread has started          #and thread is finished
        if thread._started.is_set() and not thread.is_alive(): #Thread has finished
          thread.join()
        
          self.fetch_results.pop(thread_id) #Remove request

      ######## Permit other threads temporary access to variables ########
      self.critical_lock.release()
    
    get_logger().info('Closing Monitor Thread.')

  def _start_sub_thread(self,):
    self._close_thread = False  #Used is shutdown to terminate monitor thread
    self.critical_lock = Lock() #Used for synchronizing critical points in async code
    self.monitor_thread = Thread(
      target=ParallelManager._monitor_clients, 
      args=(self,)
    )
    self.monitor_thread.daemon = True
    self.monitor_thread.start()


  def start(self, initializer=None, initargs=()):
    self.register('cleanup',           callable=self.cleanup)
    self.register('_start_sub_thread', callable=self._start_sub_thread)

    super().start(initializer, initargs)
    self._start_sub_thread()

  def cleanup(self,):
    """
      Shuts down subthreads and releases memory allocations of tasks
    """
    get_logger().info('Shutting down monitor threads.')
    self._close_thread = True
    self.monitor_thread.join()

    get_logger().info('Releasing memory allocations.')
    del self.tasks
    client_ids = list(self.clients.keys())
    for client_id in client_ids:
      client = self.clients[client_id]
      listener = client['listener']
      server_conn = client['connection']
      server_conn.close()
      listener.close()
      del self.clients[client_id]
  
  def _get_idle_clients(self,):
    idle_clients = []
    for client_id, client in self.clients.items():
      if client['alive'] and not client['busy'] and client['connection']:
        idle_clients.append(client_id)
    
    return idle_clients

  def _get_new_task_id(self):
    """
      Returns a new task_id.
      Used for internal monitoring of tasks.
    """
    task_id = self.current_task_id
    self.current_task_id += 1
    return task_id
  
  def _get_new_fetch_results_thread_id(self):
    """
      Returns a new task_id.
      Used for internal monitoring of tasks.
    """
    thread_id = self.fetch_results_thread_id
    self.fetch_results_thread_id += 1
    return thread_id
  
  def _get_new_client_id(self,):
    """
      Returns a new task_id.
      Used for internal managing of client connections.
    """
    client_id = self.current_client_id
    self.current_client_id += 1
    return client_id

  def submit(self, func, *args, timeout=None, **kwargs):
    """
      Interface for 'clients' to submit a problem and its dependencies to the 
      problem hoster, the 'server'. The server hosts this problem and loads it 
      into a self regulated task queue. 

      #Returns server's identifying task id if submission was successful. 
      #Otherwise, None. 

      # Arguments:
      func:    an exectuable function.
      args:    all arguments to be passed into func.
      kwargs:  all keyword arguments to be passed into func
      timeout: the max time in seconds to permit the function to be in 
        operation. If None, the default for the manager, as created
        by __init__(), will be used.
    """
    ######## Critical Code (Async access to variables prohibited) ########
    self.critical_lock.acquire()

    data = None
    if self.task_limit is None or len(self.tasks) < self.task_limit:
      task_id = str(self._get_new_task_id())
      self.queued_tasks.add(task_id)

      task = {
        'task_id':          task_id,
        'func':             func,
        'args':             args,
        'kwargs':           kwargs,
        'start_time':       None, 
        'running_on':       None, #client_id 
        'completed':        False,
        'result':           None,
        'timeout':          timeout or self.timeout,
        'terminated_early': False
      }
      self.tasks[task_id] = task

      data = task_id
      get_logger().info(f'Task received. Queued tasks: {len(self.queued_tasks)}.')
    
    self.critical_lock.release()
    return Packet(data)

  def monitor(self,):
    """
      Interface for client to monitor the server for active tasks
      Registers the client on the server and will permit assigning tasks to the 
      client
      
      Returns connection. The server will send packets to the client through 
      this connection when tasks are waiting in the servers task queue.
    """
    ######## Critical Code (Async access to variables prohibited) ########
    self.critical_lock.acquire()

    ip = self._address[0]
    port = get_free_port()
    address = (ip, port)
    listener = Listener(address, authkey=self._authkey)
    
    client_id = self._get_new_client_id()
    self.clients[client_id] = {
       'listener':   listener,
       'connection': None,
       'busy':       False,
       'alive':      True,
       'times_dead': 0
    }
    
    #Currently letting garbage collector manager joining this thread.
    thread = Thread(
      target=ParallelManager._connect_monitor_client,
      args=(self, client_id, listener)
    )
    thread.daemon = True
    thread.start()

    self.critical_lock.release()
    return Packet(Promise(address))
  
  @staticmethod
  def _connect_monitor_client(self, client_id, listener):
    server_conn = listener.accept()
    ip, port = listener.last_accepted
    get_logger().info(f'Client {client_id} connected at {ip}:{port}')
    
    ######## Critical Code (Async access to variables prohibited) ########
    self.critical_lock.acquire()
    
    self.clients[client_id]['connection'] = server_conn

    self.critical_lock.release()
  
  def _send_queued_task(self, client_id):
    """
      Sends queued task to client
      
      Finds a queued tasks and sends it to Client
    """
    client = self.clients[client_id]
    server_conn = self.clients[client_id]['connection']
    
    try:
      task_id = self.queued_tasks.pop() #if no task available, throws KeyError
      self.active_tasks.add(task_id)
      task = self.tasks[task_id]
      
      msg = f'Sending task {task_id} to client {client_id}.'
      get_logger().info(msg)

      needed_fields = {
        'task_id': task.get('task_id'),
        'func':    task.get('func'),
        'args':    task.get('args'),
        'kwargs':  task.get('kwargs'),
      }
      packet = Packet(needed_fields)
      server_conn.send(packet)

      msg = f'Task {task_id} sent. ' + \
        f'Queued tasks: {len(self.queued_tasks)}'
      get_logger().info(msg)

      #Update Client Status
      client['busy'] = True
      
      #Update Task Status
      modInfo = {
        'start_time': datetime.datetime.now(), 
        'running_on': client_id,
      }
      task.update(modInfo)
    except KeyError:
      pass
    except Exception as e:
      self.queued_tasks.add(task_id)
      msg = f'Error sending task {task_id} to client.\n' + \
        f'Terminating client {client_id} connection.'
      msg += f'\nException: {e}'
      self._kill_client(client_id)
      get_logger().warning(msg)

  
  def _kill_tasks(self, task_ids):
    """
      Terminates active tasks and closes manager to listening for a 
      response for those specific tasks. Called in event a timeout is reached.
      This is called to ensure tasks do not exceed their maximum duration.
    """
    for task_id in task_ids:
      task = self.tasks.get(task_id)
      if task and task['running_on']:
        client_id = task['running_on']
        #Update Client
        self._kill_client(client_id)

        #Update Task
        task['running_on'] = None
        task['completed'] = datetime.datetime.now()
        task['terminated_early'] = True

        get_logger().warning(f'Task {task_id} timed out on client {client_id}.')

        try:
          self.active_tasks.remove(task_id)
          self.completed_tasks.add(task_id)
        except Exception as e:
          get_logger().warning(f'Error occured killing tasks: {e}')
  
  def _kill_client(self, client_id):
    try:
      client = self.clients[client_id]
      client['alive'] = False
      client['busy']  = False
      client['times_dead'] += 1

      #Disconnect client
      if client['times_dead'] > self.revive_limit:
        listener = client['listener']
        server_conn = client['connection']
        server_conn.close()
        listener.close()
        del self.clients[client_id]
        get_logger().warning(f'Client {client_id} has been disconnected.')

    except KeyError: #Another thread has already killed process
      pass
  
  def _complete_task(self, task_id, result):
    task = self.tasks[task_id]
    client_id = task['running_on']

    #Update Client
    client = self.clients[client_id]
    client['busy']  = False
  

    #Update Task
    task['result']     = result
    task['running_on'] = None
    task['completed']  = datetime.datetime.now()
    get_logger().info(f'Results received for task {task_id}.')

    try:
      self.active_tasks.remove(task_id)
      self.completed_tasks.add(task_id)
    except Exception as e:
      get_logger().warning(f'Error occured completing task {task_id}: {e}')
  
  def _clear_task(self, task_id):
    """
      Removes all traces of a task being present on the server.
      Removes the task_id from all task queues and opens memory for additional 
      tasks.
    """

    for collection in [self.queued_tasks, self.active_tasks, self.completed_tasks]:
      if task_id in collection:
        collection.remove(task_id)
    
    self.tasks.pop(task_id, None)

  def get_results(self, task_ids=None, values_only=True, clear=True):
    """
      Interface for driver to request completed tasks' results

      task_ids: list of task_ids as generated by self.submit(). These are used by the 
        server to identify which task to return the results for. 
        If `None`, waits for and returns results for all tasks.
      values_only: Remove task_id before returning values. Returned answers are in 
        the order of the 'task_ids' parameter.
      clear: If True, removes task from server memory after returning results.
    """
    ######## Critical Code (Async access to variables prohibited) ########
    self.critical_lock.acquire()

    thread_id = self._get_new_fetch_results_thread_id()
    ip = self._address[0]
    port = get_free_port()
    address = (ip, port)
    listener = Listener(address, authkey=self._authkey)

    thread = Thread(
      target=ParallelManager._get_results_thread,
      args=(self, listener, task_ids, values_only, clear)
    )
    self.fetch_results[thread_id] = {
      'thread': thread,
      'listener': listener,
    }
    thread.start()

    self.critical_lock.release()

    promise = Promise(address)
    return Packet(promise)

  @staticmethod
  def _get_results_thread(self, listener, task_ids, values_only, clear):
    server_conn = listener.accept() #Accept client's connection. Send to client through this connection

    if task_ids is None:
      task_ids = list(self.tasks.keys())

    task_set = set(task_ids)
    while task_set.difference(self.completed_tasks):
      pass
    
    ######## Critical Code (Async access to variables prohibited) ########
    self.critical_lock.acquire()
    
    results = []
    for task_id in task_ids:
      task = self.tasks[task_id]
      
      result = task.get('result')
      results.append(result)
      if clear:
        self._clear_task(task_id)
    
    self.critical_lock.release()

    if values_only:
      data = results
    else:
      data = dict(zip(task_ids, results))

    packet = Packet(data)
    try:
      server_conn.send(packet)
    except:
      msg = 'Connection Error. Unable to fulfill fetch-results promise.'
      get_logger().critical(msg)
    server_conn.close()
    listener.close()


import zlib
import dill
import pickle
import socket
import datetime
from .logging import get_logger
from multiprocessing.connection import Client

def get_local_ip():
  return socket.gethostbyname(socket.gethostname())

def get_free_port():
  """
    Returns an OS determined available socket.
  """
  s = socket.socket(socket.AF_INET)
  s.bind(('localhost', 0))
  port = s.getsockname()[1]
  s.close()
  return port

class Promise:
  """
    Client side packet for waiting for incoming packages and syncing 
    subprocesses accross remote nodes.
  """

  def __init__(self, address):
    self.conn = None
    self.address = address

  def wait(self, authkey):
    """
      Connect to server at the promised address
      Then poll the underlying socket until data is returned by server.
      Once data has been received, closes the connection.

      This is done to permit client to poll sockets rather than the server.

      # Returns
      Data from the server endpoint specified by the underlying address.
    """
    if not self.conn:
      self.connect(authkey)

    while not self.conn.poll(): #Wait for data to be available
      pass

    data = self.conn.recv().unpack()
    self.conn.close()
    self.conn = None

    return data
  
  def connect(self, authkey, timeout=None):
    """
      Connects to server at the promised address.

      Returns the client side connection that can be used to communincate with 
      the server.
    """
    if not self.conn:
      
      now = datetime.datetime.now
      start_time = now()
      while self.conn is None and \
        (timeout is None or (now() - start_time).total_seconds() > timeout):
        self._connect(authkey)
    
    return self.conn
  
  def _connect(self, authkey):
    try:
      self.conn = Client(self.address, authkey=authkey)
      connected = True
    except ConnectionRefusedError:
      logger = get_logger()
      logger.warn('Promised connection was refused by server. Reattempting...')
      connected = False
    return connected

class Packet:
  """
    A basic Packet class for synchronization of data via Proxys.
    Proxies will return a Packet object. Use packet.unpack() to obtain 
    contained data.
  """

  def __init__(self,data):
    self.data = data
    self.times_compressed = 0
    self.serialize_method = None
  
  def unpack(self,):
    """
      Utility function that decompresses then returns the data stored in the 
      packet.
    """
    self.decompress()
    return self.data

  def compress(self, level=1, iterations=1, threshold=0):
    """
      Serializes and compresses `self.data` to reduce payload across Pipes
      Useful when sending data accross a Proxy or Pipe to a remote manager.

      # Arguments
      level: ZLIB compress parameter, -1 to 9 that dictates compression vs time
        efficiency. 1 is lowest compression but fastest. 9 is greatest 
        compression. 0 means no compression, -1 means to intuit what level will \
        be the most efficient for speed and memory
      iterations: How many times to compress. If None, will continuue compression
        until further compression no longer saves memory.
      threshold: Maximum size in bytes for the payload before compression is 
        deemed necessary. If the serialized payload exceeds this much 
        data, then it will also be compressed. If `None`, then will serialize, 
        but not compress.
    """
    try:
      serialized = pickle.dumps(self.data)
      self.serialize_method = 'pickle'
    except Exception:
      serialized = dill.dumps(self.data) #Can throw dill error, should do so.
      self.serialize_method = 'dill'
    
    data = serialized
    if threshold is not None and len(data) > threshold: #If packet is sufficiently large, compress

      if iterations is None: #Compress until compression adds bytes
        compressed = self._compress(serialized, level)
        while len(compressed) < len(data):
          data = compressed
          compressed = self._compress(compressed, level) #adds 1 and end that must be offset
        self.times_compressed -= 1 #offsetting to omit final compression

      elif iterations >= 1:
        compressed = serialized
        for i in range(iterations):
          compressed = self._compress(compressed, level)
        data = compressed
    
    self.data = data
  
  def _compress(self, data, level):
    compressed = zlib.compress(data, level=level)
    self.times_compressed += 1 #Should add after compress finished so try/catch can be managed elsewhere
    return compressed
  
  def decompress(self):
    """
      Identifies if deserialization or decompression is necessary. If so, 
      this function deserializes and/or decompresses the stored data.
    """
    data = self.data
    times_compressed = self.times_compressed
    for i in range(times_compressed):
      data = zlib.decompress(data)
      self.times_compressed -= 1
    
    deserialized = data
    if self.serialize_method:
      if self.serialize_method == 'pickle':
        deserialized = pickle.loads(data)
        self.serialize_method = None
      
      elif self.serialize_method == 'dill':
        deserialized = dill.loads(data)
        self.serialize_method = None
      
      else:
        msg = 'Cant Deserialize Packet. ' + \
          "Expected Serialization method to be one of [\'dill\', \'pickle\'], " + \
          f'but got {self.serialize_method}.'
        raise Exception(msg)
    
    self.data = deserialized


import datetime
from multiprocessing import Process
from .logging import get_logger
from .utils import Packet, get_local_ip
from .base import BaseDispatcher
from .managers import ParallelManager

class DistributedDispatcher(BaseDispatcher):

  def __init__(self, server_ip='127.0.0.1', port=50000, authkey=b'authkey', 
              task_limit=None, timeout=None, revive_limit=0):
    """
      Initializes a Distributed & Multicore Backend Remote Manager.
      # Arguments
      server_ip: String. IP address for Remote Server. Client machines must use 
        and be able to see this machine.
      port: Int. Port number to open for clients to interface with the manager.
      authkey: Byte string. Used to authenticate access to the manager.
      task_limit: Int. Max number of tasks for server to remember. 
        This monitors the total number of active, completed, and queued tasks.
      timeout: Default time in seconds to permit on ther server for a task. 
        If a task takes longer, the server ceases to await a response.
      revive_limit: Int. Max number of times a client can reach a timeout before 
        it is disconnected from the server.
    """

    #Denotes that a server has started on this thread. Only true on the server.
    self._started      = False
    self.port          = port
    self.authkey       = authkey
    self.server_ip     = server_ip
    self.manager_creds = (server_ip, port, authkey)
  
    # Start a shared manager server and access its queues
    self.manager = ParallelManager(
      address=(server_ip, port), 
      authkey=authkey, 
      timeout=timeout,
      task_limit=task_limit,
      revive_limit=revive_limit
    )

  def spawn_server(self, duration=None):
    """
      Initializes a server on the active thread for the alloted duration

      duration: time in seconds to leave server running. If `None`, 
      server will run and hang the current thread indefinitely.
    """
    manager = self.manager
    manager.start()
    self._started = True
    
    start_time = datetime.datetime.now()

    # server = manager.get_server()
    ip = get_local_ip()
    logger = get_logger()
    logger.info(f'Server started. {ip}:{self.port}')
    
    time_running = (datetime.datetime.now() - start_time).total_seconds()
    while duration is None or time_running <= duration:
      time_running = (datetime.datetime.now() - start_time).total_seconds()
    
    self.shutdown()

  def spawn_client(self, workers=1):
    """
      Uses the active thread to connect to the remote server.
      Sets the client to monitor the connected server for tasks. When tasks are 
      available, client will request the necessary functions and data to 
      complete, and then submit the computed result to the server.
      # Arguments
      workers: Int. How many worker processes to spawn on the client.
    """

    if workers > 1:
      processes = []
      for worker_id in range(workers):
        p = Process(
          target=type(self)._spawn_client_wrapper, 
          args=((worker_id,) + self.manager_creds),
        )
        p.start()
        processes.append(p)
      
      for process in processes:
        process.join()
    else:
      type(self)._spawn_client_wrapper(*self.manager_creds)

  @staticmethod
  def _spawn_client_wrapper(worker_id, server_ip, port, authkey):
    """
      Wrapper for multiprocessing backend to spawn clients in subprocesses.
    """
    manager = ParallelManager(address=(server_ip, port), authkey=authkey)
    manager.connect()
    
    logger = get_logger()
    logger.info(f'Worker {worker_id} connected to: {manager.address}.')
    promise = manager.monitor().unpack() #Register client with server.
    conn = promise.connect(authkey)
    while True:
      while not conn.poll(): #Wait for data to be available
        pass
      
      try:
        packet = conn.recv()
      except (ConnectionAbortedError, ConnectionResetError, EOFError):
        logger.critical(f'Worker {worker_id} was disconnected from server.')
        break
      
      task = packet.unpack()

      
      #Unpack info and compute result
      if task is not None:
        task_id   = task['task_id']
        func      = task['func']
        args      = task['args']
        kwargs    = task['kwargs']
        
        logger.info(f'Worker {worker_id} received task {task_id}.')
        
        result    = func(*args, **kwargs)

        logger.info(f'Worker {worker_id} finished task {task_id}.')
        
        response = Packet({
            'task_id': task_id,
            'result': result
        })
        conn.send(response)
  
  def shutdown(self,):
    """
      Closes monitor threads of manager and associated subprocesses
    """
    logger = get_logger()
    if self._started:
      logger.info('Shutting down server.')
      try:
        self.manager.cleanup()
        self.manager.shutdown()
      except Exception as e:
        msg = f'Error occured shutting down server: {e}'
        logger.critical(msg)

  def run(self, func, *args, timeout=None, **kwargs):
    """
      Places a task into the servers queued tasks for client completion.
      # Returns the associated task id to recover the results.
      # Arguments
      func: function to be run via clients. Needs to be a function visible to 
        the requesting machine, the server, and the clients.
      args: arguments for `func`.
      kwargs: keyword arguments for `func`.
      timeout: the max time in seconds to permit the function to be in 
        operation. If None, the default for the manager, as created
        by __init__(), will be used.
    """
    manager = self.manager
    manager.connect()
    packet = manager.submit(func, *args, timeout=timeout, **kwargs)
    task_id = packet.unpack()
    return task_id
  
  def get_results(self, task_ids=None, values_only=True, clear=True):
    """
      Gets the results to tasks associated with the passed in task_ids. 
      Hangs current thread until all the tasks are complete.

      task_ids: list of task_ids as generated by self.run(). These are used by the 
        server to identify which task to return the results for. 
        If `None`, waits for and returns results for all tasks.
      values_only: if False returns dictionary that includes the task ids with its 
        results. Otherwise, returns the values computed in order of the 
        requested task ids.
      clear: If True, removes tasks from server memory after returning results.
    """
    manager = self.manager
    manager.connect()
    promise = manager.get_results(
        task_ids, 
        values_only=values_only, 
        clear=clear).unpack()
    results = promise.wait(self.authkey)
    return results


import logging

__all__ = [
  'get_logger', 'set_logfile', 'set_loglevel', 
  'disable', 'enable'
]

LOGGER_NAME = 'blitzen'
LOGGING_FORMAT = '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
_logger = None
_log_level = logging.INFO

def get_logger():
  global _logger
  global _log_level
  if not _logger:
    _logger = logging.getLogger(LOGGER_NAME)
    formatter = logging.Formatter(
      '%(asctime)s %(name)-8s %(levelname)-8s %(message)s',
      datefmt='%Y-%m-%d %H:%M:%S'
    )
    handler = logging.StreamHandler()
    handler.setFormatter(formatter)
    _logger.addHandler(handler)
    _logger.setLevel(_log_level)

  return _logger

def set_logfile(filename):
  global _logger
  global _log_level
  if _logger:
    fh = logging.FileHandler(filename)
    fh.setLevel(_log_level)
    _logger.addHandler(fh)

def set_loglevel(level):
  global _logger
  global _log_level
  if not _logger:
    _log_level = level
    _logger.setLevel(_log_level)

def disable():
  global _logger
  if not _logger:
    _logger = get_logger()
  _logger.disabled = True

def enable():
  global _logger
  if not _logger:
    _logger = get_logger()
  _logger.disabled = False


import pathlib
from setuptools import setup, find_packages

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

setup(name='blitzen',
      version='1.0.0',
      packages=find_packages(),
      description='Multiprocessing and Distributed computing dispatcher toolkit.',
      long_description=README,
      long_description_content_type="text/markdown",
      author = 'Blake Richey',
      author_email='blake.e.richey@gmail.com',
      url='https://github.com/BlakeERichey/blitzen',
      license='MIT',
      install_requires=[
        'dill>=0.3.3',
      ],
    )